/*
    libporty - async background processes for UNIX
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

/* P_net_async_bgproc_dupfds tester redefines dup2 */
#ifndef dup2
#include <unistd.h>
#endif

int P_net_async_bgproc_dupfds(P_net_socket *parent_fds, int *child_fds, int inherit_len)
{
	int n, max, pending, bound;
	char *inherit;
	enum {
		FD_WC = 1,   /* child wants to use this fd (not yet bound) */
		FD_P = 2,    /* still used as a parent fd at the moment */
		FD_KEEP = 4  /* do not close this fd - final fd for the child */
	};

	max = 0;
	for(n = 0; n < inherit_len; n++) {
		P_net_set_blocking(parent_fds[n]);
		if (parent_fds[n] > max)
			max = parent_fds[n];
		if (child_fds[n] > max)
			max = child_fds[n];
	}
	max++;
	pending = inherit_len;

	/* inherit[] is a bitfield that describes if a given fd is a parent fd to inherit, will be an fd in the child */
	inherit = calloc(max*2, 1);
	for(n = 0; n < inherit_len; n++) {
		if (inherit[parent_fds[n]] & FD_P) {
			/* parent n referenced twice: n->x,n->y  can't do this on windows */

			fprintf(stderr, "P_net_async_bgproc_dupfds error: parent fd %d referenced twice\n", parent_fds[n]);
			return -1;
		}
		if (inherit[child_fds[n]] & FD_WC) {
			/* child n referenced twice: x->n,y->n */
			fprintf(stderr, "P_net_async_bgproc_dupfds error: child fd %d referenced twice\n", child_fds[n]);
			return -2;
		}
		inherit[parent_fds[n]] |= FD_P;
		inherit[child_fds[n]]  |= FD_WC;
		inherit[child_fds[n]]  |= FD_KEEP;
	}

	/* remove fds inherited as is (n->n) */
	for(n = 0; n < inherit_len; n++) {
		if (parent_fds[n] == child_fds[n]) {
			inherit[parent_fds[n]] &= ~FD_P;
			inherit[child_fds[n]]  &= ~FD_WC;
			pending--;
			/* fprintf(stderr, " leaving alone: %d->%d\n", parent_fds[n], child_fds[n]); */
		}
	}

	while(pending) {
		do {
			bound = 0;

			/* try to dup what can be dupped (child fd is free) */
			for(n = 0; n < inherit_len; n++) {
				if (((inherit[child_fds[n]] & FD_P) == 0) && ((inherit[child_fds[n]] & FD_WC) != 0)) {
					/* parent is not using this fd */
					dup2(parent_fds[n], child_fds[n]);
					inherit[child_fds[n]] &= ~FD_WC;
					close(parent_fds[n]);
					inherit[parent_fds[n]] &= ~FD_P;
					bound++;
					pending--;
				}
			}
		} while(bound);

		/* what is left is a circular redirection - needs to be broken somewhere
		   with a swap and the rest can be resolved with the above cycle until
		   the next circular redirection loop */
		for(n = 0; n < inherit_len; n++) {
			if ((inherit[child_fds[n]] & FD_WC) != 0) {
				/* found a candidate child, move parent out to max */
				dup2(parent_fds[n], max);
				close(parent_fds[n]);
				inherit[parent_fds[n]] &= ~FD_P;
				parent_fds[n] = max;
				inherit[parent_fds[n]] |= FD_P;
				
				max++;
				break;
			}
		}
	}

	/* TODO: there may be a set of fcntls to avoid invalid closes */

	/* close all fds which are not inherited (marked as FD_KEEP) */
	for(n = 0; n < max; n++)
		if ((inherit[n] & FD_KEEP) == 0)
			close(n);

	/* close the rest of the fds until... 1024? */
	for(;n < 1024; n++)
		close(n);

/* this free() is sort of useless since exec() will scrap the whole process image */
	free(inherit);
}

int P_net_async_bgproc_inherit(char *argv[], P_net_socket *ctrlfd, P_net_socket *parent_fds, int *child_fds, int inherit_len)
{
	pid_t pid;
	int ret, n;

	children_created++;
	pid = fork();
	if (pid < 0)
		return -128;
	if (pid == 0) {
		/* set back SIGPIPE for the child (P_net_init() has ignored it) */
		signal(SIGPIPE, SIG_DFL);

		/* rebind stdio to the child side of the socket pair */
		P_net_async_bgproc_dupfds(parent_fds, child_fds, inherit_len);
		execvp(argv[0], argv);
		/* an exit(1) is tempting, but it would cause the child to close all fds;
		   this in turn would flush FILE * buffers in the parent, causing the content
		   of the buffer to be written out twice. */
		raise(9);                                   /* try to kill -9 the failed proc */
		execlp("/bin/true", "/bin/true", NULL);     /* if that did not work, exec true - that will replace the process image so buffers are lost */
	}

	/* set up watchdog for the control socket if required */
	if (ctrlfd != NULL) {
		int cfd;
		P_net_socket inv = P_NET_INVALID_SOCKET;
		ret = P_net_local_socket(ctrlfd, &cfd, P_net_nonblock_full);
		if (ret != 0)
			return ret;

		watchdog_add(&cfd, &pid, &inv);
	}
	return 0;
}

static int watchdog_wait(P_net_socket cfd, int numpids, P_pid_t *pids, bgproc_status_t *exit_status)
{
	pid_t pid;
	int status;
	int n;

/*	fprintf(stderr, "Waiting\n");*/

	/* check what's happened */
	pid = waitpid(-1, &status, WNOHANG);

/*	fprintf(stderr, "Wait return: %d\n", pid);*/

	if (pid <= 0)
		return -1;

	/* look up pid in the database */
	for(n = 0; n < numpids; n++) {
/*		printf("CHK: %d %d\n", pids[n], pid);*/
		if (pids[n] == pid) {
			/* found the process */
			if (WIFEXITED(status)) {
				exit_status->reason = BGPROC_EXITED;
				exit_status->code   = WEXITSTATUS(status);
			}
			else  if (WIFSIGNALED(status) && ((WTERMSIG(status) == 15) || (WTERMSIG(status) == 9))) {
				exit_status->reason = BGPROC_KILLED;
				exit_status->code   = 0xFFFF;
			}
			else {
				exit_status->reason = BGPROC_ABNORMAL;
				exit_status->code   = 0xFFFF;
			}
/*			fprintf(stderr, "ok1 %d\n", n);*/
			return n;
		}
	}

/*			fprintf(stderr, "ok2\n");*/
	/* not our child */
	return -1;
}

static void watchdog_kill(P_pid_t pid, int hard)
{
	kill(pid, hard ? 9 : 15);
}

