/*  gentrex - generic tree expressions; VM to evaluate regex-like expressions
    Copyright (C) 2015  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/gentrex
    Author: gentrex@igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "gentrex.h"
#include "execute.h"
#include "gentrex.gram.h"
#include "gentrex.lex.h"



/* instruction names for dump */
static const char *gtx_inst_names[] = {
	"NODE_REF",
	"NOT",
	"TREE_CHILD",
	"TREE_CHILDREN",
	"TREE_PARENT",
	"TREE_NEXT",
	"BRPUSH",
	"AND",
	"OR",
	"ASG",
	"TAG_EQ",
	"TAG_MATCH",
	"TAG_C_OP",
	"TAG_TRUE",
	"ENTER",
	"LEAVE"
};

/********* compilation *********/
void gtx_error(gentrex_state_t *ctx, const char *msg)
{
	ctx->err_msg = msg;
}

int gtx_wrap(yyscan_t yyscanner)
{
	return 1;
}


gtx_inst_t *gtx_append(gentrex_state_t *ctx, gtx_inst_type_t it)
{
	gtx_inst_t *i;
	if (ctx->code == NULL)
		ctx->code = calloc(sizeof(gtx_code_t), 1);

	/* grow the code memory if needed */
	if (ctx->code->used >= ctx->code->alloced) {
		ctx->code->alloced += 64;
		ctx->code->code = realloc(ctx->code->code, ctx->code->alloced * sizeof(gtx_inst_t));
	}

	/* claim the next instruction and set type */
	i = &ctx->code->code[ctx->code->used];
	i->inst = it;
	i->c_op = NULL;
	ctx->code->used++;
	return i;
}

/********* execution *********/
/* push a node onto the executaiton stack of the context; grow stack if needed */
static void push(gentrex_state_t *ctx, gentrex_nodecomp_t data)
{
	if (ctx->sp >= ctx->sa) {
		ctx->sa += 64;
		ctx->stack = realloc(ctx->stack, ctx->sa * sizeof(gentrex_nodecomp_t));
	}
	ctx->stack[ctx->sp] = data;
	ctx->sp++;
}

/* pop a node from the execution stack of the context; no checks */
static void pop(gentrex_state_t *ctx, gentrex_nodecomp_t *result)
{
	*result = ctx->stack[--ctx->sp];
}

/* append a node to the end of the open list; grow the list if needed */
static void openlist_put(gtx_search_state_t *ss, gentrex_nodecomp_t ndc, int depth)
{
	if (ndc.straight == NULL)
		return;
	if (ss->ol_used >= ss->ol_alloced) {
		ss->ol_alloced += 128;
		ss->ol = realloc(ss->ol, ss->ol_alloced * sizeof(gtx_ol_t));
	}
	ss->ol[ss->ol_used].ndc   = ndc;
	ss->ol[ss->ol_used].depth = depth;
	ss->ol_used++;
}

/* get and remove next node from the beginning of the open list; returns 0
   if there are no more nodes */
static int openlist_get(gtx_search_state_t *ss, int *depth, gentrex_nodecomp_t *result)
{
	if (ss->ol_head >= ss->ol_used)
		return 0;
	*depth = ss->ol[ss->ol_head].depth;
	*result = ss->ol[ss->ol_head++].ndc;
	return 1;
}

/* access macros to make the execute code more readable */

/* top item on the stack */
#define TOP    (ctx->sp - 1)

/* current state of the search named by the current instruction */
#define ISTATE (ctx->search_state[i->arg1.range.state_idx])

/* range of the current instruction */
#define IRANGE i->arg1.range

/* whether the current instruction has ranges */
#define RANGED (i->arg1.range.state_idx >= 0)

/* invert a complementer-node */
#define INVERT(nc) \
do { \
	if (nc.straight == NULL) { \
		nc.straight = nc.inverted; \
		nc.inverted = NULL; \
	} \
	else { \
		nc.inverted = nc.straight; \
		nc.straight = NULL; \
	} \
} while(0)

/* turn nc into false (if it is not already false) */
#define FALSIFY(nc) \
do { \
	if (nc.straight != NULL) { \
		nc.inverted = nc.straight; \
		nc.straight = NULL; \
	} \
} while(0)

/* import a node as a complementer; false_inv is used as the false value
   when node is NULL */
#define IMPORT(nc, node, false_inv) \
do { \
	void *__node__ = node; \
	void *__finv__ = false_inv; \
	if (__node__ != NULL) { \
		nc.straight = __node__; \
		nc.inverted = NULL; \
	} \
	else { \
		nc.straight = NULL; \
		nc.inverted = __finv__; \
	} \
} while(0)

/* Check range of current search:
    - if we are before the first hit
      - if there is no more iterations, return INVALID
      - else try again (restore node to before_restore and jump back to before_jumpback)
    - if we are after the last interesting item, stop iterating
   NOTE: out of range doesn't mean false, but means totally INVALID/irrelevant:
   make a special FALSE that remains FALSE even when inverted */
#define CHK_RANGE(before_restore, before_jumpback) \
do { \
	if (ISTATE.depth < IRANGE.from) { \
		if (ISTATE.depth >= 0) { \
			ctx->stack[TOP] = before_restore; \
			goto before_jumpback;\
		} \
		ctx->stack[TOP].straight = ctx->stack[TOP].inverted = NULL; \
	} \
	if ((IRANGE.to >= 0) && (ISTATE.depth > IRANGE.to)) \
		ctx->stack[TOP].straight = ctx->stack[TOP].inverted = NULL; \
} while(0)

/* evaluate an expression once from a given root using the current search state */
static void *gentrex_execute_(gentrex_state_t *ctx, void *tree_root)
{
	gtx_inst_t *i;
	gentrex_nodecomp_t op1;
	int inc_succ = 0;
	void *(*tree_cb)(void *user_tree_ctx, void *node);
	gentrex_nodecomp_t NUT;

	if ((ctx->code == NULL) || (ctx->code->code == NULL) || (tree_root == NULL))
		return NULL;

	NUT.straight = tree_root;
	NUT.inverted = NULL;

	/* reset the stack, set NUT to the root */
	ctx->sp = 0;
	push(ctx, NUT);

	for(i = ctx->code->code; i - ctx->code->code < ctx->code->used; i++) {
		switch(i->inst) {
			case INST_NODE_REF:
				ctx->stack[TOP].straight = ctx->stack[TOP].inverted = NULL;
				ctx->api->map_get(ctx->api->user_map_ctx, i->arg2.str, &ctx->stack[TOP]);
				break;
			case INST_NOT:
				INVERT(ctx->stack[TOP]);
				break;
			case INST_ASG:
				ctx->api->map_set(ctx->api->user_map_ctx, i->arg2.str, ctx->stack[TOP]);
				break;

			/* non-recursive tree instructions: first child, direct parent, sibling */
			case INST_TREE_CHILD:
				tree_cb = ctx->api->child;
				goto tree_inst;
			case INST_TREE_PARENT:
				tree_cb = ctx->api->parent;
				goto tree_inst;
			case INST_TREE_NEXT:
				tree_cb = ctx->api->next;
				tree_inst:;
				op1 = ctx->stack[TOP];
				if (!RANGED) { /* not ranged: a direct relative of the node, not iterating at all */
					if (ctx->stack[TOP].straight != NULL) {
						IMPORT(ctx->stack[TOP], tree_cb(ctx->api->user_tree_ctx, ctx->stack[TOP].straight), ctx->stack[TOP].straight);
					}
					break;
				}
				if (ISTATE.depth < 0) {
					/* init: pick the first item-of-iteration */
						if (ctx->stack[TOP].straight != NULL)
							ISTATE.depth = 0;
						ISTATE.ndc = ctx->stack[TOP];
				}
				else if (ISTATE.inc) {
				tree_inst_next:;
					/* inc: pick the next item-of-iteration */
					IMPORT(ctx->stack[TOP], tree_cb(ctx->api->user_tree_ctx, ISTATE.ndc.straight), ISTATE.ndc.straight);
					ISTATE.ndc = ctx->stack[TOP];
					ISTATE.inc = 0;
					ISTATE.depth++;
					inc_succ = (ctx->stack[TOP].straight != NULL);
					if (!inc_succ)
						ISTATE.depth = -1;
				}
				else {
					/* restore */
					ctx->stack[TOP] = ISTATE.ndc;
				}
				CHK_RANGE(op1, tree_inst_next);
				break;

			/* recursive tree instruction: children is a BFS */
			case INST_TREE_CHILDREN:
				op1 = ctx->stack[TOP];
				if (!RANGED) {
					assert(!"broken gentrex grammar: / is /{1}");
				}
				if (ISTATE.depth < 0) {
						if (ctx->stack[TOP].straight != NULL)
							ISTATE.depth = 0;
						ISTATE.ndc = ctx->stack[TOP];
				}
				else if (ISTATE.inc) {
					ISTATE.inc = 0;
				tree_children_next:;
					if (ISTATE.depth == 0) {
						/* select the first child */
						IMPORT(ctx->stack[TOP],  ctx->api->child(ctx->api->user_tree_ctx, ISTATE.ndc.straight), ISTATE.ndc.straight);
						ISTATE.ndc = ctx->stack[TOP];
						ISTATE.depth++;
						/* if there's a node here and ISTATE.depth+1 is still in range, put children on open list */
						if ((ctx->stack[TOP].straight != NULL) && ((IRANGE.to < 0) || (ISTATE.depth < IRANGE.to))) {
							gentrex_nodecomp_t tmp;
							IMPORT(tmp, ctx->api->child(ctx->api->user_tree_ctx, ISTATE.ndc.straight), ISTATE.ndc.straight);
							openlist_put(&ISTATE, tmp, ISTATE.depth+1);
						}
						inc_succ = (ctx->stack[TOP].straight != NULL);
						if (!inc_succ)
							ISTATE.depth = -1;
					}
					else {
						/* select the next sibling on the same level */
						if (ISTATE.ndc.straight != NULL)
							IMPORT(ctx->stack[TOP], ctx->api->next(ctx->api->user_tree_ctx, ISTATE.ndc.straight), ISTATE.ndc.straight);
						else
							FALSIFY(ctx->stack[TOP]);
						ISTATE.ndc = ctx->stack[TOP];
						if (ctx->stack[TOP].straight == NULL) {
							openlist_get(&ISTATE, &ISTATE.depth, &ctx->stack[TOP]);
							ISTATE.ndc = ctx->stack[TOP];
						}
						if (ctx->stack[TOP].straight != NULL) {
							gentrex_nodecomp_t tmp;
							IMPORT(tmp, ctx->api->child(ctx->api->user_tree_ctx, ctx->stack[TOP].straight), ctx->stack[TOP].straight);
							openlist_put(&ISTATE, tmp, ISTATE.depth+1);
						}
						inc_succ = (ctx->stack[TOP].straight != NULL);
						if (!inc_succ)
							ISTATE.depth = -1;
					}
				}
				else {
					/* restore */
					ctx->stack[TOP] = ISTATE.ndc;
				}

				CHK_RANGE(op1, tree_children_next);
				break;

			/* branching */
			case INST_BRPUSH:
				if (i->arg1.brtype == INST_AND) {
					if (ctx->stack[TOP].straight == NULL) { /* when left side is false, skip the right side */
						i = i->arg2.branch - 1;
						push(ctx, ctx->stack[TOP]);
						break;
					}
					else {
						push(ctx, NUT);
					}
				}
				else if (i->arg1.brtype == INST_OR) {
					if (ctx->stack[TOP].straight != NULL) { /* when left side is true, skip the right side */
						gentrex_nodecomp_t tmp;
						i = i->arg2.branch - 1;
						tmp = ctx->stack[TOP];
						INVERT(tmp);
						push(ctx, tmp);
						break;
					}
					else
						push(ctx, NUT);
				}
				else {
					assert(!"internal error: unknonw brtype");
				}
				break;
			case INST_AND:
				pop(ctx, &op1);
				if (!(op1.straight && ctx->stack[TOP].straight))
					FALSIFY(ctx->stack[TOP]);
				break;
			case INST_OR:
				pop(ctx, &op1);
				if (op1.straight || ctx->stack[TOP].straight) {
					/* keep the one that is true */
					if (ctx->stack[TOP].straight == NULL)
						ctx->stack[TOP] = op1;
				}
				else
					FALSIFY(ctx->stack[TOP]);
				break;

			/* tag matching */
			case INST_TAG_EQ:
				if (ctx->stack[TOP].straight != NULL)
					if (!(ctx->api->tag_eq(ctx->api->user_tree_ctx, ctx->stack[TOP].straight, i->match_cookie, i->arg1.tagn, i->arg2.str)))
						FALSIFY(ctx->stack[TOP]);
				break;
			case INST_TAG_MATCH:
				if (ctx->stack[TOP].straight != NULL)
					if ((ctx->api->tag_match == NULL) || (!(ctx->api->tag_match(ctx->api->user_tree_ctx, ctx->stack[TOP].straight, i->match_cookie, i->arg1.tagn, i->arg2.str))))
						FALSIFY(ctx->stack[TOP]);
				break;
			case INST_TAG_C_OP:
				if (ctx->stack[TOP].straight != NULL)
					if ((ctx->api->tag_c_op == NULL) || (!(ctx->api->tag_c_op(ctx->api->user_tree_ctx, ctx->stack[TOP].straight, i->match_cookie, i->arg1.tagn, i->arg2.str, i->c_op))))
						FALSIFY(ctx->stack[TOP]);
				break;
			case INST_TAG_TRUE:
				/* return true if NUT is a valid node: translates to a no-op */
				break;
			case INST_ENTER:
				push(ctx, NUT);
				NUT = ctx->stack[TOP-1];
				push(ctx, NUT);
				/* stack: ... old_root NUT
				                        ^TOP
				*/
				break;
			case INST_LEAVE:
				/* stack: ... old_root NUT result
				                            ^TOP
				*/
				NUT = ctx->stack[TOP-2];
				ctx->stack[TOP-2] = ctx->stack[TOP];
				pop(ctx, &op1);
				pop(ctx, &op1);
				/* stack: ... result
				               ^TOP
				*/
				break;
		}
	}
	assert(TOP == 0);
	return ctx->stack[TOP].straight;
}

/* increase search state; sort of a numerical increment on a number
   that has each digit represented in a different base. Rollover of a digit
   happens when the given level couldn't find any more nodes. Increasing
   a higher digit may present different NUT for the lower digits. */
int gentrex_inc_search(gentrex_state_t *ctx)
{
	int n;

	/* find the "lowest-digit" (LSB) state that can be increased */
	for(n = ctx->search_state_len-1; n >= 0; n--) {
		if ((ctx->search_state[n].depth != -1) && (ctx->search_state[n].ndc.straight != NULL)) {
			/* .depth != -1 means it's already running, .node != NULL means it's not finished yet */
			ctx->search_state[n].inc = 1;

			/* after increasing the given state, reset all lower states so they can be increased too */
			for(n++; n < ctx->search_state_len; n++) {
				ctx->search_state[n].depth = -1;
				ctx->search_state[n].ndc.straight = ctx->search_state[n].ndc.inverted = NULL;
				ctx->search_state[n].ol_head = ctx->search_state[n].ol_used = 0; /* paranoia: make sure a previous open list doesn't leak in */
			}
			return 1;
		}
	}

	/* even the MSB rolled around: the search is over */
	return 0;
}

/* loop expression evaluation and search state increase until 
   we find a match (evaluation returns non-FALSE) or increasing
   the search state rolled over globally. */
void *gentrex_execute(gentrex_state_t *ctx, void *tree_root)
{
	if (ctx->err_msg != NULL)
		return NULL;

	while(!ctx->reached_end) {
		void *res;
	
		res = gentrex_execute_(ctx, tree_root);
		ctx->reached_end = !gentrex_inc_search(ctx);
		if (res != NULL)
			return res;
	}
	return NULL;
}

void gentrex_reset(gentrex_state_t *ctx)
{
	int n;
	for(n = 0; n < ctx->search_state_len; n++) {
		ctx->search_state[n].inc   = 0;
		ctx->search_state[n].depth = -1;
		ctx->search_state[n].ndc.straight = ctx->search_state[n].ndc.inverted = NULL;
		if (ctx->search_state[n].ol != NULL)
			free(ctx->search_state[n].ol);
		ctx->search_state[n].ol_used = 0;
		ctx->search_state[n].ol_head = 0;
		ctx->search_state[n].ol_alloced = 0;
	}
	ctx->reached_end = 0;
}


gentrex_state_t *gentrex_alloc(const gentrex_api_t *api, const char *script, const char **error_msg)
{
	gentrex_state_t *ctx;
	struct yy_buffer_state *buff;

	ctx = calloc(sizeof(gentrex_state_t), 1);

	ctx->api = api;

	gtx_lex_init(&ctx->scanner);
	buff = gtx__scan_string(script, ctx->scanner);
	gtx_parse(ctx);
	gtx__delete_buffer(buff, ctx->scanner);
	gtx_lex_destroy(ctx->scanner);

	if (error_msg != NULL)
		*error_msg = ctx->err_msg;

	if (ctx->err_msg != NULL) {
		ctx->search_state_len = 0;
		gentrex_free(ctx);
		return NULL;
	}

	if (ctx->search_state_len > 0)
		ctx->search_state = calloc(sizeof(gtx_search_state_t), ctx->search_state_len);
	gentrex_reset(ctx);
	return ctx;
}

void gentrex_free(gentrex_state_t *ctx)
{
	gentrex_reset(ctx);
	if (ctx->stack != NULL)
		free(ctx->stack);
	if (ctx->search_state != NULL)
		free(ctx->search_state);
	if (ctx->code != NULL) {
		gtx_inst_t *i;
		int n;
		for(i = ctx->code->code, n = 0; n < ctx->code->used; n++, i++) {
			switch(i->inst) {
				case INST_ASG:
				case INST_NODE_REF:
					free(i->arg2.str);
					break;
				case INST_TAG_C_OP:
				case INST_TAG_EQ:
				case INST_TAG_MATCH:
					if ((ctx->api->op_cookie_free != NULL) && (i->match_cookie != NULL))
						ctx->api->op_cookie_free(ctx->api->user_tree_ctx, i->inst == INST_TAG_MATCH, i->arg1.tagn, i->arg2.str, i->match_cookie, i->c_op);
					free(i->arg1.tagn);
					free(i->arg2.str);
					if (i->c_op != NULL)
						free(i->c_op);
					break;
				default:; /* suppress warning */
			}
		}
		if (ctx->code->code != NULL) {
			free(ctx->code->code);
		}
		free(ctx->code);
	}

	if (ctx->match_c_op != NULL)
		free(ctx->match_c_op);

	free(ctx);
}

/********* diagnostics *********/
void gentrex_dump(gentrex_state_t *ctx, void *tree_root)
{
	gtx_inst_t *i;
	int n;

	if ((ctx->code == NULL) || (ctx->code->code == NULL))
		return;

	for(i = ctx->code->code, n = 0; n < ctx->code->used; n++, i++) {
		/* print address and instruction */
		printf("%03d  %-16s ", i - ctx->code->code, gtx_inst_names[i->inst]);

		/* print operands */
		switch(i->inst) {
			case INST_ASG:
			case INST_NODE_REF:
				printf("name=%s", i->arg2.str);
				break;
			case INST_TREE_CHILD:
			case INST_TREE_CHILDREN:
			case INST_TREE_PARENT:
			case INST_TREE_NEXT:
				if (i->arg1.range.state_idx >= 0) {
					printf("range(%d)=[%d..", i->arg1.range.state_idx, i->arg1.range.from);
					if (i->arg1.range.to < 0)
						printf("infinite]");
					else
						printf("%d]", i->arg1.range.to);
				}
				break;
			case INST_NOT:
			case INST_AND:
			case INST_OR:
			case INST_TAG_TRUE:
			case INST_ENTER:
			case INST_LEAVE:
				/* no operand */
				break;
			case INST_BRPUSH:
				printf("branch=%03d type=%s", i->arg2.branch - ctx->code->code, gtx_inst_names[i->arg1.brtype]);
				break;
			case INST_TAG_EQ:
			case INST_TAG_MATCH:
			case INST_TAG_C_OP:
				printf("name=%s pattern=%s c_op=%s", i->arg1.tagn, i->arg2.str, i->c_op);
				break;
		}
		printf("\n");
	}
}
