typedef struct c99c_loc_file_s {
	char *fn;              /* file name, as present in #include */
	char *path;            /* full path (note search paths) */
} c99c_loc_file_t;

typedef struct c99c_loc_s {
	c99c_loc_file_t *file;      /* host file */
	int soffs, eoffs;      /* byte offsets from the start of the file (start/end) counting from 0 */
	int sline, scol;       /* start line/col counting from 1 */
	int eline, ecol;       /* end line/col counting from 1 */

	/* NOTE: update c99c_loc_cpy_start()/c99c_loc_cpy_end() when new fields are inserted */

	int offs; /* last seen byte offset in the file */

#ifdef PARSER_NALLOC_DEBUG
	int yline;
#endif

	int macro_level;          /* macro expansion nesting level; 0 means not in macro */
	int is_local;             /* whether the location is local (directly in the target file, not in included file) */

	int cpp_line;
	const char *cpp_file;

	c99c_node_t *twin_parent;      /* parent in the twin tree: for a grammar tree node, this is the token node */
	int twin_parent_uid;      /* for the duration of loading nodes from a file: valid only as long as twin_parent is NULL */
	int twin_parent_uid_orig; /* a copy of the original (load-time) value */
} c99c_loc_t;

/* a string with location: for saving token start info */
typedef struct c99c_strloc_s {
	char *str;
	c99c_loc_t loc;
} c99c_strloc_t;

/* print location to c99c_loc_buff, using c99c_loc_format (return a pointer to c99c_loc_buff) */
const char *c99c_loc_print(c99tree_cc_t *ctx, const c99c_node_t *node, char *c99c_loc_buff, int c99c_loc_buff_size);

/* convert a location found in in, put result in c99c_loc_scan_buff, use c99c_loc_format */
c99c_loc_t *c99c_loc_scan(c99tree_cc_t *ctx, const char *in, c99c_loc_t *c99c_loc_scan_buff);

/* Open a new file, set it as current c99c_loc_file, pushing the old current on
   top of the stack; hash files by path */
void c99c_loc_file_open(c99tree_cc_t *ctx, const char *fn, const char *path);

/* Close current c99c_loc_file, pop loc from the stack, make that the current one */
void c99c_loc_file_close(c99tree_cc_t *ctx);

/* New location file name; check if it's the currently open file. If not,
   decide whether it's an include or the end of the current input */
void c99c_loc_switch_file(c99tree_cc_t *ctx, const char *fn, const char *path);

/* Copy current loc end fields to start fields, advance the end
   fields as needed using token string; returns token to shorten source
   of the .l */
const char *c99c_loc_update_end(c99tree_cc_t *ctx, const char *token);

/* Copy location struct */
void c99c_loc_cpy(c99c_loc_t *dst, const c99c_loc_t *src);

/* Copy location not related to start/end (file, cpp) */
void c99c_loc_cpy_misc(c99c_loc_t *dst, const c99c_loc_t *src);

/* Copy token-start fields of a location struct */
void c99c_loc_cpy_start(c99c_loc_t *dst, const c99c_loc_t *src);

/* Copy token-end fields of a location struct */
void c99c_loc_cpy_end(c99c_loc_t *dst, const c99c_loc_t *src);

/* set dst location to start..end; if start or end is NULL, no modification
   is made on that field of dst. */
void c99c_loc_set(c99c_loc_t *dst, const c99c_loc_t *start, const c99c_loc_t *end);

/* macro shorthands for the most common case: copy between node locations */
#define C99Nloc_cpy(ctx, ndst, nsrc)        c99c_loc_cpy(&((ndst)->loc), &((nsrc)->loc))
#define C99Nloc_cpy_start(ctx, ndst, nsrc)  c99c_loc_cpy_start(&((ndst)->loc), &((nsrc)->loc))
#define C99Nloc_cpy_end(ctx, ndst, nsrc)    c99c_loc_cpy_end(&((ndst)->loc), &((nsrc)->loc))


/* return a cpp location file name - have unique pointer to each file name */
const char *c99c_loc_cpp_file_lookup(c99tree_cc_t *ctx, const char *fn);

/* increase location line counters and set column back to 0 */
void c99c_loc_newline(c99tree_cc_t *ctx);

/* Delete the location files database and stack */
void c99c_loc_destroy(c99tree_cc_t *ctx);


