/*
    c99tree - parse tree utility
    Copyright (C) 2010-2015  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "c99tree-cc.h"
#include "tree_print.h"
#include "genht/htsi.h"
#include "genht/hash.h"

/* use the true versions in this file so that the original locations are inherited */
#undef C99Nalloc
#undef C99Nalloc_wloc
#undef C99Nalloc_tok
#undef C99Nclone
#undef C99Nop
#undef C99Nop_tok

const char *c99c_node_type_names[] = {
"CALL", "STATEMENT", "TYPE", "DECL", "FUNCDEF", "ARGLIST", "ARG",
"BLOCK", "CLASS", "TYPEOF", "ATTRLIST", "ATTR", "PTR", "INDEX",
"QUALIFIER", "ID", "ELLIPSIS", "ENUM", "INIT", "STRUCT", "BITS", "ASM",
"LABEL", "CONDITION", "THEN", "ELSE", "OP", "ICONST",
"FCONST", "SCONST", "SCONSTLIST", "DEBUG", "UNION", "PRAGMA", "TYPECUSTOM",
"TYPESPEC", "TYPENAME", "TYPEDEF",  "EXPRLIST", "EXPRBRACKET", "CSE",
"COMMENT", "INITLIST",  "GRAMMAR_TREE", "TOKEN_TREE", "TOKEN_STR",
"TOKEN_PPDIR", "GENTREX", "GENTREX_RESULT", "GENTREX_MAP", "SYMLINK", NULL};

htsi_t *c99c_node_type_hash = NULL;

c99c_node_type_t C99Ntype(c99tree_cc_t *ctx, const char *typenam)
{
	const htsi_entry_t *ent;
	ent = (const htsi_entry_t *)htsi_getentry(c99c_node_type_hash, (htsi_key_t)typenam);
	if (ent == NULL)
		return T_INVALID;
	return (c99c_node_type_t)ent->value;
}

/* --- hash --- */
static int keyeq(char *a, char *b)
{
	return !strcmp(a, b);
}

static unsigned int hash(char *key)
{
	return murmurhash(key, strlen(key));
}


void c99c_tree_init(c99tree_cc_t *ctx)
{
	int n;

	c99c_node_type_hash = htsi_alloc(hash, keyeq);

	for(n = 0; n < T_INVALID; n++)
		htsi_set(c99c_node_type_hash, (htsi_key_t)c99c_node_type_names[n], n);
}

void c99c_tree_uninit(c99tree_cc_t *ctx)
{
	htsi_free(c99c_node_type_hash);
}

void C99Nloc_cpy_subtree(c99tree_cc_t *ctx, c99c_node_t *dest, c99c_node_t *child)
{
	c99c_node_t *n;
	int m;

	for(n = child; n != NULL; n = n->next) {

		m = 0;
		if (n->loc.soffs >= 0) {
			if ((dest->loc.soffs < 0) || (n->loc.soffs < dest->loc.soffs)) {
				C99Nloc_cpy_start(ctx, dest, n);
			 m = 1;
			}
		}

		if (n->loc.eoffs > dest->loc.eoffs) {
			C99Nloc_cpy_end(ctx, dest, n);
			m = 1;
		}
		
		if (m)
			c99c_loc_cpy_misc(&dest->loc, &n->loc);

		if (n->children != NULL)
			C99Nloc_cpy_subtree(ctx, dest, n->children);
	}
}

c99c_node_t *C99Nalloc_wloc(c99tree_cc_t *ctx, c99c_node_type_t type, char *name, c99c_node_t *child, const c99c_loc_t *start, const c99c_loc_t *end YLINEARG)
{
	c99c_node_t *n;

	n = malloc(sizeof(c99c_node_t));
	n->type = type;
	n->uid  = ++ctx->node_uid_counter;
	n->name = name;
	n->next = NULL;
	n->parent = NULL;
	n->children = child;
	n->redefinable = 0;
	n->user_data = NULL;

	if ((start == NULL) || (end == NULL)) {
		/* if either start or end is uncertain, use current location and child */
		c99c_loc_cpy(&(n->loc), ctx->loc_current);
		if (child != NULL) {
			n->loc.eoffs = -1;
			C99Nloc_cpy_subtree(ctx, n, child);
		}
	}

	/* copy end first, so start overrides the misc section */
	if (end != NULL) {
		c99c_loc_cpy_misc(&n->loc, end);
		c99c_loc_cpy_end(&n->loc, end);
	}

	if (start != NULL) {
		c99c_loc_cpy_misc(&n->loc, start);
		c99c_loc_cpy_start(&n->loc, start);
	}

#ifdef PARSER_NALLOC_DEBUG
	n->loc.yline = yline;
#endif

	return n;
}

c99c_node_t *C99Nalloc(c99tree_cc_t *ctx, c99c_node_type_t type, char *name, c99c_node_t *child YLINEARG)
{
	return C99Nalloc_wloc(ctx, type, name, child, NULL, NULL YLINEPARM);
}

c99c_node_t *C99Nalloc_tok(c99tree_cc_t *ctx, c99c_node_type_t type, c99c_strloc_t *token, c99c_node_t *child YLINEARG)
{
	c99c_node_t *n;

	n = malloc(sizeof(c99c_node_t));
	n->type = type;
	n->uid  = ++ctx->node_uid_counter;
	n->name = token->str;
	n->next = NULL;
	n->parent = NULL;
	n->children = child;
	n->redefinable = 0;

	c99c_loc_cpy(&(n->loc), &(token->loc));

#ifdef PARSER_NALLOC_DEBUG
	n->loc.yline = yline;
#endif

	return n;
}

c99c_node_t *C99Nappend(c99tree_cc_t *ctx, c99c_node_t *dest, c99c_node_t *src)
{
	c99c_node_t *odest;

	odest = dest;

	for(; dest->next != NULL; dest = dest->next) ;
	dest->next = src;

	return odest;
}

static void C99Nprint_(c99tree_cc_t *ctx, void **nprint_ctx, const char *prefix, c99c_node_t *node, int local_only)
{
	char *pr2;
	c99c_node_t *n;
	int l, r;

	if (!local_only || node->loc.is_local)
		r = ctx->Nprint_cb(ctx, nprint_ctx, prefix, node, PS_OPENING);
	else
		r = 0;

	if ((r == 0) && (node->children != NULL)) {
		l = strlen(prefix);
		pr2 = malloc(l + 2);
		*pr2 = *prefix;
		memcpy(pr2+1, prefix, l+1);
		for(n = node->children; n != NULL; n = n->next)
			C99Nprint_(ctx, nprint_ctx, pr2, n, local_only);
		free(pr2);
	}

	if (!local_only || node->loc.is_local)
		ctx->Nprint_cb(ctx, nprint_ctx, prefix, node, PS_CLOSING);
}

void C99Nprint(c99tree_cc_t *ctx, void **nprint_ctx, const char *prefix, c99c_node_t *node, int local_only)
{
	c99c_node_t *n;
	for(n = node; n != NULL; n = n->next)
		C99Nprint_(ctx, nprint_ctx, prefix, n, local_only);
}

c99c_node_t *C99Nop(c99tree_cc_t *ctx, const char *op, c99c_node_t *o1, c99c_node_t *o2, c99c_loc_t *start, c99c_loc_t *end YLINEARG)
{
	c99c_node_t *res;

	res = C99Nalloc_wloc(ctx, T_OP, c99c_strclone(op), o1, start, end YLINEPARM);
	if (o2 != NULL)
		C99Nappend(ctx, o1, o2);
	return res;
}

c99c_node_t *C99Nop_tok(c99tree_cc_t *ctx, const c99c_strloc_t *op, c99c_node_t *o1, c99c_node_t *o2 YLINEARG)
{
	c99c_node_t *res;

	res = C99Nalloc_wloc(ctx, T_OP, c99c_strclone(op->str), o1, &op->loc, &op->loc YLINEPARM);
	if (o2 != NULL)
		C99Nappend(ctx, o1, o2);
	return res;
}

c99c_node_t *C99Nclone(c99tree_cc_t *ctx, const c99c_node_t *node, invalidate_loc_t invalidate_loc YLINEARG)
{
	c99c_node_t **prev, *t, *ret;
	const c99c_node_t *n;

	if (node == NULL)
		return NULL;

	prev = &ret;
	for(n = node; n != NULL; n = n->next) {
		int cpy;
		*prev = t = C99Nalloc(ctx, n->type, n->name == NULL ? NULL : c99c_strclone(n->name), C99Nclone(ctx, n->children, invalidate_loc YLINEPARM) YLINEPARM);

		switch(invalidate_loc) {
			case INV_LOC_ALL:        cpy = 0; break;
			case INV_LOC_NONE:       cpy = 1; break;
			case INV_LOC_ALL_BUT_ID: cpy = t->type != T_ID; break;
		}
		if (cpy)
			c99c_loc_cpy(&t->loc, &ctx->loc_invalid);
		prev = &(t->next);
	}
	*prev = NULL;
	return ret;
}

void C99Nfree(c99tree_cc_t *ctx, c99c_node_t *node)
{
	c99c_node_t *next;
	c99c_node_t *n;

	for(n = node; n != NULL; n = next) {
		if (n->children != NULL)
			C99Nfree(ctx, n->children);
		next = n->next;
		if (n->name != NULL)
			free(n->name);
		free(n);
	}
}

c99c_node_t *C99Nlast(c99tree_cc_t *ctx, c99c_node_t *node)
{
	while(node->next != NULL) node = node->next;
	return node;
}

void c99c_name_all_siblings(c99tree_cc_t *ctx, c99c_node_t *node, const char *name)
{
	for(; node != NULL; node = node->next) {
		if (node->name != NULL)
			free(node->name);
		node->name = c99c_strclone(name);
	}
}

void C99Nreplace(c99tree_cc_t *ctx, c99c_node_t *oldn, c99c_node_t *newn)
{
	c99c_node_t *p = NULL, *n;
	if (oldn->parent == NULL)
		return;
	for(n = oldn->parent->children; n != NULL; p = n, n = n->next) {
		if (n == oldn) {
			if (p == NULL) {
				if (newn == NULL)
					oldn->parent->children = oldn->next;
				else
					oldn->parent->children = newn;
			}
			else {
				if (newn == NULL)
					p->next = oldn->next;
				else
					p->next = newn;
			}
			break;
		}
	}
	if (newn != NULL) {
		newn->next = oldn->next;
		newn->parent = oldn->parent;
	}
	oldn->next = NULL;
	oldn->parent = NULL;
}

void C99Ninsert_after(c99tree_cc_t *ctx, c99c_node_t *after, c99c_node_t *newn)
{
	newn->next = after->next;
	after->next = newn;
	newn->parent = after->parent;
}


void C99Nunlink(c99tree_cc_t *ctx, c99c_node_t *node)
{
	C99Nreplace(ctx, node, NULL);
}

